<?php
namespace EMS\PortalFuncionario\Geral\Services;

use Zeedhi\Framework\Controller\Crud;
use Doctrine\ORM\EntityManager;

use EMS\PortalFuncionario\Util\{ DateUtil, NovoCodigo, Distancia, HcmServices };

use EMS\PortalFuncionario\Geral\Repository\{MarcacaoRepository, EstruturaRepository};
use EMS\PortalFuncionario\Util\Utilities;
use EMS\PortalFuncionario\Geral\Services\ParamOrgService;

// MENSAGENS
define('messages', [
    'ERRO_LOCALIZACAO_INCORRETA'    => 'Você não está na área definida como de marcação do ponto!',
    'ERRO_ESTRUT_GERENC_EXIGE_FOTO' => 'A estrutura gerencial exige a foto na marcação!',
    'ERRO_PONTO_ACESSO_NÃO_ENCONTRADO' => 'A estrutura do vínculo não possui ponto de acesso cadastrado!'
]);

class MarcacaoService extends Crud {
    private const pontoAcessoDefaults = array(
        'VRDISTANCIAMAXREF' => 0,
        'IDOBRIGAFOTO' => 'N',
        'DSLATITUDE' => 0,
        'DSLONGITUDE' => 0
    );

    /** @var EntityManager */
    protected $entityManager;

    /** @var MarcacaoRepository */
    protected $marcacaoRepository;

    /** @var EstruturaRepository */
    protected $estruturaRepository;

    /** @var Utilities */
    protected $utilities;

    /** @var HcmServices */
    protected $hcmServices;

    /** @var ParamOrgService */
    protected $paramOrgService;

    public function __construct(
        EntityManager       $entityManager,
        MarcacaoRepository  $marcacaoRepository,
        EstruturaRepository $estruturaRepository,
        Utilities           $utilities,
        HcmServices         $hcmServices,
        ParamOrgService     $paramOrgService
    ) {
        $this->entityManager            = $entityManager;
        $this->marcacaoRepository       = $marcacaoRepository;
        $this->estruturaRepository      = $estruturaRepository;
        $this->utilities                = $utilities;
        $this->hcmServices              = $hcmServices;
        $this->paramOrgService          = $paramOrgService;
    }

    public function registraMarcacaoIntervalo(
        array $row
    ): array {
        $route = 'registraMarcacaoIntervalo';

        $nrOrg = (int) $row['NRORG'];

        $nrColetorBatida = $this->paramOrgService->getNrColetorBatida($nrOrg);

        $row['NRCOLETOR'] = $nrColetorBatida;

        $params = [
            'requestType' => 'Row',
            'row' => $row,
        ];

        $result = $this->hcmServices->send($route, $params, $nrOrg);
        $setMarcacaoStatus = $result['dataset']['setMarcacaoIntervalo'];

        return $setMarcacaoStatus;
    }

    public function registraMarcacaoAlmoco(
        array $row
    ): array {
        $route = 'registraMarcacaoAlmoco';

        $nrOrg = (int) $row['NRORG'];

        $nrColetorBatida = $this->paramOrgService->getNrColetorBatida($nrOrg);

        $row['NRCOLETOR'] = $nrColetorBatida;

        $params = [
            'requestType' => 'Row',
            'row' => $row,
        ];

        $result = $this->hcmServices->send($route, $params, $nrOrg);
        $setMarcacaoStatus = $result['dataset']['setMarcacaoAlmoco'];

        return $setMarcacaoStatus;
    }

    public function registraOcorrenciaPonto(
        array $row
    ): array {
        $route = 'registraOcorrenciaPonto';

        $nrOrg = (int) $row['NRORG'];

        $nrColetorBatida = $this->paramOrgService->getNrColetorBatida($nrOrg);

        $row['NRCOLETOR'] = $nrColetorBatida;

        $params = [
            'requestType' => 'Row',
            'row' => $row,
        ];

        $result = $this->hcmServices->send($route, $params, $nrOrg);
        $setMarcacaoStatus = $result['dataset']['setOcorrenciaPonto'];

        return $setMarcacaoStatus;
    }

    public function getValidacaoMarcacaoDistancia(
        string $nrvinculom,
        string $nrorg,
        string $dslatitude,
        string $dslongitude,
               $pontoAcesso = null
    ) {
        $pontoAcessoValidado = $this->validaPontoAcesso($nrorg, $nrvinculom, $pontoAcesso);

        // Validações do ponto de acesso
        $marcacaoCerca = $pontoAcessoValidado['VRDISTANCIAMAXREF'];

        // Validação do local
        $latitudeReferencia = $pontoAcessoValidado['DSLATITUDE'];
        $longitudeReferencia = $pontoAcessoValidado['DSLONGITUDE'];

        // Valida se deve checar a distância entre o ponto de referência e o ponto de batida do ponto
        $distancia = 0;
        
        if ($marcacaoCerca != 0) {
            $distancia = Distancia::getDistanceBetweenLocations(
                $latitudeReferencia,
                $longitudeReferencia,
                $dslatitude,
                $dslongitude
            );
        }

        // Confirma se a marcação está sendo realizada em uma zona válida
        $distanciaValida = $distancia <= $marcacaoCerca;
        $diferencaDistancia = $distancia - $marcacaoCerca;
        $diferencaDistancia = round($diferencaDistancia, 3);

        if ($distanciaValida) {
            $diferencaDistancia = 0;
        }

        return array(
            'DISTANCIAVALIDA'       => $distanciaValida,
            'DISTANCIADIFERENCA'    => $diferencaDistancia,
            'IDOBRIGAFOTO'          => $pontoAcessoValidado['IDOBRIGAFOTO'],
            'VRDISTANCIAMAXREF'     => $pontoAcessoValidado['VRDISTANCIAMAXREF'],
            'DSLATITUDE'            => $pontoAcessoValidado['DSLATITUDE'],
            'DSLONGITUDE'           => $pontoAcessoValidado['DSLONGITUDE']
        );
    }

    public function getValidacaoMarcacaoFoto(
        string $nrvinculom,
        string $nrorg,
               $existeFotoRequisicao,
               $pontoAcesso = null
    ) {
        $pontoAcessoValidado = $this->validaPontoAcesso($nrorg, $nrvinculom, $pontoAcesso);

        $idObrigaFoto = $pontoAcessoValidado['IDOBRIGAFOTO'];
        $estruturaExigeFotoMarcacao = $idObrigaFoto == 'S';

        $obrigacaoCumprida = false;
        $enviaFoto = false;

        if ($estruturaExigeFotoMarcacao && $existeFotoRequisicao) {
            $obrigacaoCumprida = true;
            $enviaFoto = true;
        } else if (!$estruturaExigeFotoMarcacao) {
            $obrigacaoCumprida = true;
            $enviaFoto = false;
        }

        return array(
            'ENVIAFOTO' => $enviaFoto,
            'OBRIGACAOCUMPRIDA' => $obrigacaoCumprida,
        );
    }

    public function novaMarcacao(
        string $nrvinculom,
        string $nrorg,
        string $cdoperador,
        string $dslatitude,
        string $dslongitude,
               $dspathfoto
    ) : array {
        
        try {
            // Horário base
            $horarioBase = $this->marcacaoRepository->getHorarioBase($nrvinculom, $nrorg, $cdoperador);

            $entradaHorarioBase = $horarioBase['HRENTRADA'];
            $saidaHorarioBase = $horarioBase['HRSAIDA'];

            $dataAtual = new \DateTime();
            $horaAtual = (int) $dataAtual->format('G');
            $minutoAtual = (int) $dataAtual->format('i');
        
            $horaEntradaHorarioBase = explode(':', $entradaHorarioBase)[0];
            $minutoEntradaHorarioBase = explode(':', $entradaHorarioBase)[1];
        
            $horaSaidaHorarioBase = explode(':', $saidaHorarioBase)[0];
            $minutoSaidaHorarioBase = explode(':', $saidaHorarioBase)[1];
        
            $entradaHorarioBaseDateTime = new \DateTime();
            $entradaHorarioBaseDateTime->setTime($horaEntradaHorarioBase, $minutoEntradaHorarioBase);
            $entradaHorarioBase = (int) $entradaHorarioBaseDateTime->format('G');
        
            $saidaHorarioBaseDateTime = new \DateTime();
            $saidaHorarioBaseDateTime->setTime($horaSaidaHorarioBase, $minutoSaidaHorarioBase);
            $saidaHorarioBase = (int) $saidaHorarioBaseDateTime->format('G');

            // Hora marcação do ponto
            $dthoramarcacaoponto = $dataAtual->format('d/m/Y G:i:s');

            // Dados estrutura
            $dadosEmpresa = $this->estruturaRepository->getDadosEstruturaLegal($nrvinculom, $nrorg);

            $cdcnpjestrut = $dadosEmpresa['CDCNPJESTRUT'];
            $nmrazsocestrut = $dadosEmpresa['NMRAZSOCESTRUT'];
            $cdcaepfestrut = $dadosEmpresa['CDCAEPFESTRUT'];
            $cdceiestrut = $dadosEmpresa['CDCEIESTRUT'];

            // Ticket autenticação
            $ticketAut = $this->getMarcacaoTicketAut($nrvinculom, $nrorg, $dataAtual);

            // Marcação desprezada
            $idmarcdesprezada = null;
            if($horaAtual >= $entradaHorarioBase && $horaAtual <= $saidaHorarioBase || $horaAtual == $entradaHorarioBase-1 && $minutoAtual >= 55) {
                $idmarcdesprezada     = "N";
            } else if ($horaAtual < $entradaHorarioBase) { //Se estiver antes do horário base
                $idmarcdesprezada     = "S";
            } else if ($horaAtual > $saidaHorarioBase) { //Se estiver depois do horário base
                $idmarcdesprezada     = "S";
            }
        
            $dadosMarcacao =  array(
                "IDENQUADRAHORARIO" => "N",
                "IDDIAOCORRENCIA"   => "FC",
                "DSJUSTIFICATIVA"   => "",
                "NRCOLETOR"         => 24,
                "IDREFERINTERVALO"  => "",
                "CDCNPJESTRUT"      => $cdcnpjestrut,
                "NMRAZSOCESTRUT"    => $nmrazsocestrut,
                "CDCAEPFESTRUT"     => $cdcaepfestrut,
                "CDCEIESTRUT"       => $cdceiestrut,
                "CDTICKETAUT"       => $ticketAut,
                "DSLATITUDE"        => $dslatitude,
                "DSLONGITUDE"       => $dslongitude,
                "DSPATH"            => $dspathfoto,
                "IDATIVO"           => "S",
                "NRORGINCLUSAO"     => $nrorg,
                "CDOPERINCLUSAO"    => $cdoperador,
                "NRVINCULOM"        => $nrvinculom,
                "NRORG"             => $nrorg,
                "DTHRMARCACAOPONTO" => $dthoramarcacaoponto,
                "DTHRAJUSTADA"      => $dthoramarcacaoponto,
                "IDMARCDESPREZADA"  => $idmarcdesprezada
            );
        
            return $dadosMarcacao;
        } catch (\Exception $e) {
            throw $e;
        }
    }

    public function getMarcacaoTicketAut($nrvinculom, $nrorg, $data = null) {
        if (empty($data)) {
            $data = new \DateTime();
        }

        if (!is_string($nrorg)) {
            $nrorg = strval($nrorg);
        }

        $nrorg = str_pad($nrorg, 4, "0", STR_PAD_LEFT);

        return $nrorg . $nrvinculom . $data->format('dmYGisu');
    }

    public function getPontoAcesso($nrorg, $nrvinculom) {
        $pontoAcesso = $this->estruturaRepository->getPontoAcesso($nrorg, $nrvinculom);

        if (empty($pontoAcesso)) {
            throw new \Exception(constant('messages')['ERRO_PONTO_ACESSO_NÃO_ENCONTRADO']);
        }
        
        return $pontoAcesso;
    }

    private function validaPontoAcesso($nrorg, $nrvinculom, $pontoAcesso = null) {
        if ($pontoAcesso == null) {
            $pontoAcesso = $this->getPontoAcesso($nrorg, $nrvinculom);
        }

        $vrdistanciamaxrefDefault = self::pontoAcessoDefaults['VRDISTANCIAMAXREF'];
        $idobrigafotoDefault = self::pontoAcessoDefaults['IDOBRIGAFOTO'];
        $dslatitudeDefault = self::pontoAcessoDefaults['DSLATITUDE'];
        $dslongitudeDefault = self::pontoAcessoDefaults['DSLONGITUDE'];

        $pontoAcesso['VRDISTANCIAMAXREF'] = !empty($pontoAcesso['VRDISTANCIAMAXREF']) ? $pontoAcesso['VRDISTANCIAMAXREF'] : $vrdistanciamaxrefDefault;
        $pontoAcesso['IDOBRIGAFOTO'] = !empty($pontoAcesso['IDOBRIGAFOTO']) ? $pontoAcesso['IDOBRIGAFOTO'] : $vrdistanciamaxrefDefault;
        $pontoAcesso['DSLATITUDE'] = !empty($pontoAcesso['DSLATITUDE']) ? $pontoAcesso['DSLATITUDE'] : $vrdistanciamaxrefDefault;
        $pontoAcesso['DSLONGITUDE'] = !empty($pontoAcesso['DSLONGITUDE']) ? $pontoAcesso['DSLONGITUDE'] : $vrdistanciamaxrefDefault;

        return $pontoAcesso;
    }

    public function getParesMarcacoesDia(
        int $nrOrg,
        int $nrVinculo,
        string $dia
    ) : array {

        $params = [
            'requestType' => 'Row',
            'row' => [
                'NRVINCULOM' => $nrVinculo,
                'DIA'       => $dia
            ]
        ];

        $result = $this->hcmServices->send('getParesMarcacoesDia', $params, $nrOrg);
        $paresMarcacoes = $result['dataset']['getParesMarcacoesDia'];

        return $paresMarcacoes;
    }

    public function saveParesMarcacao(
        int     $nrOrg,
        int     $nrVinculo,
        array   $dia,
        string  $hrEntrada,
        string  $hrSaida,
        string  $idDiaInicio,
        string  $idDiaFim,
        ?int    $nrMotivoJust = null,
        ?string $justificativa = null
    ): array {
        $nrColetorFuncionario = $this->paramOrgService->getNrColetorFuncionario($nrOrg);

        $params = [
            'requestType' => 'Row',
            'row' => [
                'NRVINCULOM'        => $nrVinculo,
                'DIA'               => $dia,
                'HRENTRADA'         => $hrEntrada,
                'HRSAIDA'           => $hrSaida,
                'DIA_INICIO'        => $idDiaInicio,
                'DIA_FIM'           => $idDiaFim,
                'NRCOLETOR'         => $nrColetorFuncionario,
                'NRMOTIVOJUST'      => $nrMotivoJust,
                'JUSTIFICATIVA'     => $justificativa,
            ],
        ];

        $result = $this->hcmServices->send('saveParesMarcacao', $params, $nrOrg);
        $nrMarcacoes = $result['dataset']['saveParesMarcacao']['nrMarcacoes'];

        return $nrMarcacoes;
    }

    public function removeParesMarcacao(
        int     $nrOrg,
        int     $nrVinculo,
        array   $dia,
        string  $hrEntrada,
        string  $hrSaida
    ) : array {

        $params = [
            'requestType' => 'Row',
            'row' => [
                'NRVINCULOM'        => $nrVinculo,
                'DATE'              => $dia,
                'HRENTRADA'         => $hrEntrada,
                'HRSAIDA'           => $hrSaida
            ]
        ];

        $result = $this->hcmServices->send('removeParesMarcacao', $params, $nrOrg);
        $nrMarcacoes = $result['dataset']['marcacoesRemovidas'];

        return $nrMarcacoes;
    }

    public function removeMarcacao(
        int     $nrOrg,
        int     $nrVinculo,
        string  $nrMarcacaoponto
    ) : array {

        $params = [
            'requestType' => 'Row',
            'row' => [
                'NRVINCULOM'        => $nrVinculo,
                'NRMARCACAOPONTO'   => $nrMarcacaoponto
            ]
        ];

        $result = $this->hcmServices->send('removeMarcacao', $params, $nrOrg);
        $nrMarcacoes = $result['dataset']['marcacoesRemovidas'];

        return $nrMarcacoes;
    }

    public function editParesMarcacao(
        int     $nrOrg,
        int     $nrVinculo,
        string  $hrentrada,
        string  $hrsaida,
        array   $dates,
        string  $novahrentrada,
        string  $novahrsaida
    ) : array {

        $nrColetorFuncionario = $this->paramOrgService->getNrColetorFuncionario($nrOrg);

        $params = [
            'requestType' => 'Row',
            'row' => [
                'NRVINCULOM'        => $nrVinculo,
                'DATE'              => $dates,
                'HRENTRADA'         => $hrentrada,
                'HRSAIDA'           => $hrsaida,
                'NOVAHRENTRADA'     => $novahrentrada,
                'NOVAHRSAIDA'       => $novahrsaida,
                'NRCOLETOR'         => $nrColetorFuncionario,
            ]
        ];

        $result = $this->hcmServices->send('editParesMarcacao', $params, $nrOrg);
        $nrMarcacoes = $result['dataset']['editParesMarcacao'];

        return $nrMarcacoes;
    }

    /**
     *  Rotina que retorna a hora atual do servidor PHP
     */
    public function getServerTime() {

        $dataAtual = new \DateTime();

        $serverDateTime = [
            'DATE' => $dataAtual->format('d/m/Y'),
            'TIME' => $dataAtual->format('G:i:s')
        ];

        return $serverDateTime;
    }

    /**
     *  Rotina que retorna a hora atual da base de dados
     */
    public function getDatabaseTime() {
        $sqlDateTime = $this->marcacaoRepository->getSqlTime();

        $dataBaseDateTime = [
            'DATE' => $sqlDateTime['DATE_DMY'],
            'TIME' => $sqlDateTime['TIME_H24']
        ];

        return $dataBaseDateTime;
    }

    public function fixaEntradaSaida(
        int     $nrOrg,
        int     $nrVinculo,
        string  $date,
        string  $nrMarcacaoEntrada,
        string   $nrMarcacaoSaida
    ): array {
        $route = 'fixaEntradaSaida';

        $params = [
            'requestType' => 'Row',
            'row' => [
                'NRVINCULOM'        => $nrVinculo,
                'DATE'              => $date,
                'NRMARCACAOENTRADA' => $nrMarcacaoEntrada,
                'NRMARCACAOSAIDA'   => $nrMarcacaoSaida
            ]
        ];

        $result = $this->hcmServices->send($route, $params, $nrOrg);
        $status = $result['dataset'][$route];

        return $status;
    }

    public function getMarcacoesPontoDiaOpcoesFixarMarcacao(
        int     $nrOrg,
        int     $nrVinculo,
        string  $date
    ): array {
        $route = 'getMarcacoesPontoDiaOpcoesFixarMarcacao';

        $params = [
            'requestType' => 'Row',
            'row' => [
                'NRVINCULOM'        => $nrVinculo,
                'DATE'              => $date
            ]
        ];

        $result = $this->hcmServices->send($route, $params, $nrOrg);
        $marcacoes = $result['dataset'][$route];

        return $marcacoes;
    }

    public function reverterEntradaSaida(
        int     $nrOrg,
        int     $nrVinculo,
        string  $date
    ): bool {
        $route = 'reverterEntradaSaida';

        $params = [
            'requestType' => 'Row',
            'row' => [
                'NRVINCULOM'        => $nrVinculo,
                'DATE'              => $date
            ]
        ];

        $result = $this->hcmServices->send($route, $params, $nrOrg);
        $success = $result['dataset'][$route]['success'];

        return $success;
    }

    public function getDetalheAcompanhamento(
        int   $nrOrg,
        int   $nrVinculo,
        array $dates
    ): array {
        $route = 'getDetalheAcompanhamento';

        $params = [
            'requestType' => 'Row',
            'row' => [
                'NRVINCULOM'    => $nrVinculo,
                'DATAAPURACAO'  => $dates,
            ]
        ];

        $response = $this->hcmServices->send($route, $params, $nrOrg);
        $marcacoes = $response['dataset'][$route];

        return $marcacoes;
    }

    public function sincronizarMarcacoesOffline(
        int $nrOrg,
        int $nrVinculo,
        array $offlineClockIns
    ): array {
        $route = 'sincronizarMarcacoesOffline';

        $params = [
            'requestType' => 'Row',
            'row' => [
                'NRVINCULOM' => $nrVinculo,
                'offlineClockIns' => $offlineClockIns,
            ]
        ];

        $response = $this->hcmServices->send($route, $params, $nrOrg);
        $success = $response['dataset'][$route];

        return $success;
    }

    public function isMomentoForaHorario(
        int $nrOrg,
        int $nrVinculo
    ): array {
        $route = 'isMomentoForaHorario';

        $params = [
            'requestType' => 'Row',
            'row' => [
                'NRVINCULOM' => $nrVinculo,
            ]
        ];

        $response = $this->hcmServices->send($route, $params, $nrOrg);
        $dataSet = $response['dataset'][$route];

        return $dataSet;
    }
}
